@@ -68,6 +68,14 @@ sub parse_received_headers {
   $self->{num_relays_untrusted} = 0;
   $self->{relays_untrusted_str} = '';
 
+  $self->{relays_internal} = [ ];
+  $self->{num_relays_internal} = 0;
+  $self->{relays_internal_str} = '';
+
+  $self->{relays_external} = [ ];
+  $self->{num_relays_external} = 0;
+  $self->{relays_external_str} = '';
+
   $self->{num_relays_unparseable} = 0;
 
   # now figure out what relays are trusted...
@@ -135,6 +143,12 @@ sub parse_received_headers {
       $in_internal = 0;
     }
 
+    # note: you can't be in internal networks, but not be in a trusted 
+    # net. (bug 4760)
+    if ($in_internal && !$in_trusted) {
+      $in_trusted = 1;
+    }
+
 # OK, infer the trusted/untrusted handover, if we don't have real info.
 # Here's the algorithm used (taken from Dan's mail):
 # 
@@ -262,12 +276,22 @@ sub parse_received_headers {
       push (@{$self->{relays_untrusted}}, $relay);
       $self->{allow_fetchmail_markers} = 0;
     }
+
+    if ($in_internal) {
+      push (@{$self->{relays_internal}}, $relay);
+    } else {
+      push (@{$self->{relays_external}}, $relay);
+    }
   }
 
   $self->{relays_trusted_str} = join(' ', map { $_->{as_string} }
                     @{$self->{relays_trusted}});
   $self->{relays_untrusted_str} = join(' ', map { $_->{as_string} }
                     @{$self->{relays_untrusted}});
+  $self->{relays_internal_str} = join(' ', map { $_->{as_string} }
+                    @{$self->{relays_internal}});
+  $self->{relays_external_str} = join(' ', map { $_->{as_string} }
+                    @{$self->{relays_external}});
 
   # drop the temp PerMsgStatus object
   delete $self->{dns_pms};
@@ -282,21 +306,31 @@ sub parse_received_headers {
   if ($self->{msg}->can ("delete_header")) {
     $self->{msg}->delete_header ("X-Spam-Relays-Trusted");
     $self->{msg}->delete_header ("X-Spam-Relays-Untrusted");
+    $self->{msg}->delete_header ("X-Spam-Relays-Internal");
+    $self->{msg}->delete_header ("X-Spam-Relays-External");
  
     if ($self->{msg}->can ("put_metadata")) {
       $self->{msg}->put_metadata ("X-Spam-Relays-Trusted",
 			$self->{relays_trusted_str});
       $self->{msg}->put_metadata ("X-Spam-Relays-Untrusted",
 			$self->{relays_untrusted_str});
+      $self->{msg}->put_metadata ("X-Spam-Relays-Internal",
+			$self->{relays_internal_str});
+      $self->{msg}->put_metadata ("X-Spam-Relays-External",
+			$self->{relays_external_str});
     }
   }
 
   # be helpful; save some cumbersome typing
   $self->{num_relays_trusted} = scalar (@{$self->{relays_trusted}});
   $self->{num_relays_untrusted} = scalar (@{$self->{relays_untrusted}});
+  $self->{num_relays_internal} = scalar (@{$self->{relays_internal}});
+  $self->{num_relays_external} = scalar (@{$self->{relays_external}});
 
   dbg("metadata: X-Spam-Relays-Trusted: ".$self->{relays_trusted_str});
   dbg("metadata: X-Spam-Relays-Untrusted: ".$self->{relays_untrusted_str});
+  dbg("metadata: X-Spam-Relays-Internal: ".$self->{relays_internal_str});
+  dbg("metadata: X-Spam-Relays-External: ".$self->{relays_external_str});
 }
 
 sub lookup_all_ips {
@@ -633,8 +667,10 @@ sub parse_received_line {
     # Some others of the numerous qmail patches out there can also add variants of their own
     #
     # Received: from 211.245.85.228  (EHLO ) (211.245.85.228) by mta232.mail.scd.yahoo.com with SMTP; Sun, 25 Jan 2004 00:24:37 -0800
-    if (/^from \S+( \((?:HELO|EHLO) \S*\))? \((\S+\@)?\[?${IP_ADDRESS}\]?\)( \(envelope-sender <\S+>\))? by \S+( \(.+\))* with (.* )?(SMTP|QMQP)/) {
-
+    #
+    # bug 4813: make sure that the line doesn't have " id " after the
+    # protocol since that's a sendmail line and not qmail ...
+    if (/^from \S+( \((?:HELO|EHLO) \S*\))? \((\S+\@)?\[?${IP_ADDRESS}\]?\)( \(envelope-sender <\S+>\))? by \S+( \(.+\))* with (.* )?(SMTP|QMQP)(?! id )/ ) {
        if (/^from (\S+) \((?:HELO|EHLO) ([^ \(\)]*)\) \((\S*)\@\[?(${IP_ADDRESS})\]?\)( \(envelope-sender <\S+>\))? by (\S+)/) {
          $rdns = $1; $helo = $2; $ident = $3; $ip = $4; $by = $6;
        }
@@ -732,7 +768,7 @@ sub parse_received_line {
     # Received: from cabbage.jmason.org [127.0.0.1]
     # by localhost with IMAP (fetchmail-5.9.0)
     # for jm@localhost (single-drop); Thu, 13 Mar 2003 20:39:56 -0800 (PST)
-    if (/^from (\S+) \[(${IP_ADDRESS})\] by (\S+) with \S+ \(fetchmail/) {
+    if (/^from (\S+) (?:\[(${IP_ADDRESS})\] )?by (\S+) with \S+ \(fetchmail/) {
       $self->found_pop_fetcher_sig();
       return;		# skip fetchmail handovers
     }
@@ -1222,6 +1258,8 @@ sub found_pop_fetcher_sig {
   if ($self->{allow_fetchmail_markers}) {
     dbg("received-header: found fetchmail marker, restarting parse");
     $self->{relays_trusted} = [ ];
+    $self->{relays_internal} = [ ];
+    $self->{relays_external} = [ ];
   } else {
     dbg("received-header: found fetchmail marker outside trusted area, ignored");
   }
